"""Interaction potentials."""

import torch

from .. import stateutils


class PedPedPotential(torch.nn.Module):
    """Ped-ped interaction potential based on distance b.

    v0 is in m^2 / s^2.
    sigma is in m.
    """
    delta_t_step = 0.4

    def __init__(self, v0=2.1, sigma=0.3):
        super().__init__()
        self.v0 = v0
        self.sigma = sigma

    def b(self, r_ab, speeds, desired_directions):
        """Calculate b."""
        speeds_b = speeds.unsqueeze(0)
        speeds_b_abc = speeds_b.unsqueeze(2)  # abc = alpha, beta, coordinates
        e_b = desired_directions.unsqueeze(0)

        in_sqrt = (
            self.norm_r_ab(r_ab)
            + self.norm_r_ab(r_ab - self.delta_t_step * speeds_b_abc * e_b)
        )**2 - (self.delta_t_step * speeds_b)**2

        # torch.diagonal(in_sqrt)[:] = 0.0  # protect forward pass
        in_sqrt = torch.clamp(in_sqrt, min=1e-8)
        out = 0.5 * torch.sqrt(in_sqrt)
        # torch.diagonal(out)[:] = 0.0  # protect backward pass

        return out

    def value_b(self, b):
        """Value of potential parametrized with b."""
        return self.v0 * torch.exp(-b / self.sigma)

    def value_r_ab(self, r_ab, speeds, desired_directions):
        """Value of potential explicitely parametrized with r_ab."""
        b = self.b(r_ab, speeds, desired_directions)
        return self.value_b(b)

    @staticmethod
    def r_ab(state):
        """Construct r_ab using broadcasting."""
        r = state[:, 0:2]
        r_a0 = r.unsqueeze(1)
        r_0b = r.unsqueeze(0).detach()  # detach others
        r_ab = r_a0 - r_0b
        torch.diagonal(r_ab)[:] = 0.0  # detach diagonal gradients
        return r_ab

    def forward(self, state):
        speeds = stateutils.speeds(state).detach()
        desired_directions = stateutils.desired_directions(state).detach()
        return self.value_r_ab(self.r_ab(state), speeds, desired_directions)

    def grad_r_ab_finite_difference(self, state, delta=1e-3):
        """Compute gradient wrt r_ab using finite difference differentiation."""
        r_ab = self.r_ab(state[:, 0:2])
        speeds = stateutils.speeds(state)
        desired_directions = stateutils.desired_directions(state)

        dx = torch.tensor([[[delta, 0.0]]])
        dy = torch.tensor([[[0.0, delta]]])

        v = self.value_r_ab(r_ab, speeds, desired_directions)
        dvdx = (self.value_r_ab(r_ab + dx, speeds, desired_directions) - v) / delta
        dvdy = (self.value_r_ab(r_ab + dy, speeds, desired_directions) - v) / delta

        # remove gradients from self-intereactions
        torch.diagonal(dvdx)[:] = 0.0
        torch.diagonal(dvdy)[:] = 0.0

        return torch.stack((dvdx, dvdy), dim=-1)

    def grad_r_ab(self, state):
        """Compute gradient wrt r_ab using autograd."""
        speeds = stateutils.speeds(state).detach()
        desired_directions = stateutils.desired_directions(state).detach()
        r_ab = self.r_ab(state)
        return self.grad_r_ab_(r_ab, speeds, desired_directions)

    def grad_r_ab_(self, r_ab, speeds, desired_directions):
        """Compute gradient wrt r_ab using autograd."""
        def compute(r_ab):
            return self.value_r_ab(r_ab, speeds, desired_directions)

        with torch.enable_grad():
            vector = torch.ones(r_ab.shape[0:2], requires_grad=False)
            _, r_ab_grad = torch.autograd.functional.vjp(
                compute, r_ab, vector,
                create_graph=True, strict=True)

        return r_ab_grad

    @staticmethod
    def norm_r_ab(r_ab):
        """Norm of r_ab.

        Special treatment of diagonal terms for backpropagation.

        Without this treatment, backpropagating through a norm of a
        zero vector gives nan gradients.
        """
        out = torch.linalg.norm(r_ab, ord=2, dim=2, keepdim=False)
        torch.diagonal(out)[:] = 0.0
        return out


class PedPedPotentialWall(PedPedPotential):
    """Ped-ped interaction potential based on distance b.

    v0 is in m^2 / s^2.
    sigma is in m.
    """
    delta_t_step = 0.4

    def __init__(self, sigma=0.3, w=0.1):
        super().__init__()
        self.sigma = sigma
        self.w = w

    def value_b(self, b):
        """Value of potential parametrized with b."""
        return torch.exp(-(b - self.sigma) / self.w)


class PedPedPotentialMLP(PedPedPotential):
    """Ped-ped interaction potential."""

    def __init__(self, *, hidden_units=5, small_init=False, dropout_p=0.0):
        super().__init__()

        lin1 = torch.nn.Linear(1, hidden_units)
        lin2 = torch.nn.Linear(hidden_units, 1)

        # initialize
        if small_init:
            torch.nn.init.normal_(lin1.weight, std=0.03)
            torch.nn.init.normal_(lin1.bias, std=0.03)
            torch.nn.init.normal_(lin2.weight, std=0.03)
            torch.nn.init.normal_(lin2.bias, std=0.03)

        if dropout_p == 0.0:
            self.mlp = torch.nn.Sequential(
                lin1, torch.nn.Softplus(),
                lin2, torch.nn.Softplus(),
            )
        else:
            self.mlp = torch.nn.Sequential(
                lin1, torch.nn.Softplus(), torch.nn.Dropout(dropout_p),
                lin2, torch.nn.Softplus(),
            )

    def value_b(self, b):
        """Calculate value given b."""
        b = torch.clamp(b, max=100.0)
        return self.mlp(b.view(-1, 1)).view(b.shape)



import torch


class PedSpacePotential(torch.nn.Module):
    """Pedestrian-space interaction potential.

    space is a list of numpy arrays containing points of boundaries.

    u0 is in m^2 / s^2.
    r is in m
    """

    def __init__(self, space, u0=10, r=0.2):
        super().__init__()
        self.space = space or []
        self.u0 = u0
        self.r = r

    def value_r_aB(self, r_aB):
        """Compute value parametrized with r_aB."""
        return self.u0 * torch.exp(-1.0 * torch.norm(r_aB, dim=-1) / self.r)

    def r_aB(self, state):
        """r_aB"""
        if not self.space:
            return torch.zeros((state.shape[0], 0, 2))

        r_a = state[:, 0:2].unsqueeze(1)
        closest_i = [
            torch.argmin(torch.norm(r_a - B.unsqueeze(0), dim=-1), dim=1)
            for B in self.space
        ]
        closest_points = torch.transpose(
            torch.stack([B[i] for B, i in zip(self.space, closest_i)]),
            0, 1)  # index order: pedestrian, boundary, coordinates
        return r_a - closest_points

    def forward(self, state):
        return self.value_r_aB(self.r_aB(state))

    def grad_r_aB(self, state, delta=1e-3):
        """Compute gradient wrt r_aB using finite difference differentiation."""
        r_aB = self.r_aB(state)

        dx = torch.tensor([[[delta, 0.0]]])
        dy = torch.tensor([[[0.0, delta]]])

        v = self.value_r_aB(r_aB)
        dvdx = (self.value_r_aB(r_aB + dx) - v) / delta
        dvdy = (self.value_r_aB(r_aB + dy) - v) / delta

        return torch.stack((dvdx, dvdy), dim=-1)



"""Field of view computation."""

import math
import torch


class FieldOfView:
    """Compute field of view prefactors.

    The field of view angle twophi is given in degrees.
    out_of_view_factor is C in the paper.
    """
    out_of_view_factor = 0.5

    def __init__(self, twophi=200.0):
        self.cosphi = math.cos(twophi / 2.0 / 180.0 * math.pi)

    def __call__(self, e, f):
        """Weighting factor for field of view.

        e is rank 2 and normalized in the last index.
        f is a rank 3 tensor.
        """
        cosphi_l = torch.einsum('aj,abj->ab', (e, f))
        in_sight = cosphi_l > torch.linalg.norm(f, ord=2, dim=-1) * self.cosphi

        out = torch.full_like(cosphi_l, self.out_of_view_factor)
        out[in_sight] = 1.0
        torch.diagonal(out)[:] = 0.0
        return out



import math

import numpy as np
import torch

from . import potentials
from .simulator import Simulator


class Circle:
    def __init__(self, ped_ped=None, **kwargs):
        self.ped_ped = ped_ped or potentials.PedPedPotential(2.1)
        self.simulator = Simulator(ped_ped=self.ped_ped, **kwargs)

    def generate(self, n, *, steps_per_scenario=21):
        # ped0 always left to right
        speed0 = 0.7 + 0.4 * torch.rand(1).item()
        ped0 = np.array([-5.0, 0.0, speed0, 0.0, 6.0, 0.0])

        generator_initial_states = []
        # Minimum 15degree angle (to avoid pedestrians starting on top of each other).
        # 15 degrees is about 0.25.
        for theta in 0.25 + torch.rand(n) * (2.0 * math.pi - 0.5):
            # ped1 at a random angle with +/-20% speed variation
            c, s = np.cos(theta), np.sin(theta)
            r = np.array([[c, -s], [s, c]])
            ped1 = np.concatenate((
                np.matmul(r, ped0[0:2]),
                np.matmul(r, ped0[2:4] * (0.7 + 0.4 * torch.rand(1).item())),
                np.matmul(r, ped0[4:6]),
            ))
            generator_initial_states.append(
                self.simulator.normalize_state(np.stack((ped0, ped1)))
            )

        with torch.no_grad():
            return [
                self.simulator.run(initial_state, steps_per_scenario)
                for initial_state in generator_initial_states
            ]


class ParallelOvertake:
    def __init__(self, ped_ped=None, *, b_center=0.0, **kwargs):
        self.ped_ped = ped_ped or potentials.PedPedPotential(2.1)
        self.b_center = b_center
        self.simulator = Simulator(ped_ped=self.ped_ped, **kwargs)

    def generate(self, n):
        # ped0 always left to right
        speed0 = 0.7 + 0.4 * torch.rand(1).item()
        ped0 = [-5.0, 0.0, speed0, 0.0, 6.0, 0.0]  # or tau=, 0.5

        generator_initial_states = []
        for b in 0.75 * torch.randn(n) + self.b_center:
            # with 20% speed variation
            speed = 1.2 + 0.2 * torch.rand(1).item()
            ped1 = [-7.0, b, speed, 0.0, 7.0, b]  # or tau=, 0.1

            state = np.array([ped0, ped1])
            state = self.simulator.normalize_state(state)
            generator_initial_states.append(state)

        with torch.no_grad():
            return [
                self.simulator.run(initial_state, 21)
                for initial_state in generator_initial_states
            ]


class ParallelAvoidance:
    def __init__(self, ped_ped=None, *, b_center=0.0, **kwargs):
        self.ped_ped = ped_ped or potentials.PedPedPotential(2.1)
        self.b_center = b_center
        self.simulator = Simulator(ped_ped=self.ped_ped, **kwargs)

    def generate(self, n):
        # ped0 always left to right
        speed0 = 0.7 + 0.4 * torch.rand(1).item()
        ped0 = [-5.0, 0.0, speed0, 0.0, 6.0, 0.0]  # or tau=, 0.5

        generator_initial_states = []
        for b in 0.75 * torch.randn(n) + self.b_center:
            # with 20% speed variation
            speed = 1.2 + 0.2 * torch.rand(1).item()
            ped1 = [7.0, b, -speed, 0.0, -7.0, b]  # or tau=, 0.1

            state = np.array([ped0, ped1])
            state = self.simulator.normalize_state(state)
            generator_initial_states.append(state)

        with torch.no_grad():
            return [
                self.simulator.run(initial_state, 21)
                for initial_state in generator_initial_states
            ]



"""Utility functions for plots and animations."""

from contextlib import contextmanager

import numpy as np
import torch

try:
    import matplotlib.pyplot as plt
    import matplotlib.animation as mpl_animation
except ImportError:
    plt = None
    mpl_animation = None


@contextmanager
def track_canvas(image_file=None, show=True, tight_layout=True, **kwargs):
    """Generic matplotlib context."""
    fig, ax = plt.subplots(**kwargs)

    def format_ax(this_ax):
        if isinstance(this_ax, np.ndarray):
            for a in this_ax:
                format_ax(a)
            return

        this_ax.grid(linestyle='dotted')
        this_ax.set_aspect(1.0, 'datalim')
        this_ax.set_axisbelow(True)
        this_ax.set_xlabel('$x_1$ [m]')
        this_ax.set_ylabel('$x_2$ [m]')

    format_ax(ax)

    yield ax

    if tight_layout:
        fig.set_tight_layout(True)
    if image_file:
        fig.savefig(image_file, dpi=200)
    if show:
        plt.show()
    plt.close(fig)


@contextmanager
def canvas(image_file=None, show=True, **kwargs):
    """Generic matplotlib context."""
    fig, ax = plt.subplots(**kwargs)

    yield ax

    fig.set_tight_layout(True)
    if image_file:
        fig.savefig(image_file, dpi=200)
    if show:
        plt.show()
    plt.close(fig)


@contextmanager
def animation(n, movie_file=None, writer=None, **kwargs):
    """Context for animations."""
    fig, ax = plt.subplots(**kwargs)
    fig.set_tight_layout(True)
    ax.grid(linestyle='dotted')
    ax.set_aspect(1.0, 'datalim')
    ax.set_axisbelow(True)

    context = {'fig': fig, 'ax': ax, 'update_function': None}
    yield context

    ani = mpl_animation.FuncAnimation(fig, context['update_function'], range(n))
    if movie_file:
        ani.save(movie_file, writer=writer)
    fig.show()
    plt.close(fig)


# pylint: disable=too-many-locals
def state_animation(ax, all_states, *,
                    delta_t=0.4, movie_file=None, writer=None, show_speed=False):
    pedestrian_actors = []
    text_actors = []

    for ped in range(all_states.shape[1]):
        speed = np.linalg.norm(all_states[0, ped, 2:4])
        radius = 0.2 + speed / 2.0 * 0.3
        p = plt.Circle(all_states[0, ped, 0:2], radius=radius,
                       facecolor='black' if all_states[0, ped, 6] > 0 else 'white',
                       edgecolor='black', zorder=10.0)
        pedestrian_actors.append(p)
        ax.add_patch(p)
        if show_speed:
            t = ax.text(0.1, 0.9 - 0.08 * ped, '', transform=ax.transAxes)
            text_actors.append(t)

    def update(i):
        for ped, p in enumerate(pedestrian_actors):
            # p.set_data(all_states[i:i+5, ped, 0], all_states[i:i+5, ped, 1])
            p.center = all_states[i, ped, 0:2]
            speed = np.linalg.norm(all_states[i, ped, 2:4])
            p.set_radius(0.2 + speed / 2.0 * 0.3)
            if show_speed:
                acc = np.linalg.norm(all_states[i, ped, 4:6])
                text_actors[ped].set_text(
                    f'$v_{ped} = {speed:.2f}m/s, a_{ped} = {acc:.2f}m/s^2$')
        return pedestrian_actors + text_actors

    ani = mpl_animation.FuncAnimation(
        ax.get_figure(), update,
        frames=len(all_states), interval=delta_t * 1000.0, blit=True)
    if movie_file:
        ani.save(movie_file, writer=writer, dpi=200)
    return ani


def states(ax, states, *, labels=None, monochrome=False, **kwargs):  # pylint: disable=redefined-outer-name
    states = np.asarray(states)

    if monochrome:
        kwargs['color'] = 'black'

    for ped in range(states.shape[1]):
        x = states[:, ped, 0]
        y = states[:, ped, 1]
        label = f'ped {ped}'
        if labels:
            label = labels[ped]
        tracks = ax.plot(x, y, '-o', label=label, markersize=2.5, markeredgewidth=0, **kwargs)

        marker_color = tracks[0].get_color()
        marker_alpha = tracks[0].get_alpha()
        marker_zorder = tracks[0].get_zorder()
        marker_edgewidth = 0
        marker_edgecolor = marker_color
        if monochrome and states[0, ped, 2] < 0.0:
            marker_color = 'white'
            marker_edgewidth = 1
        ax.plot(states[0, ped:ped + 1, 0], states[0, ped:ped + 1, 1],
                'o', color=marker_color,
                alpha=marker_alpha,
                markeredgewidth=marker_edgewidth,
                markeredgecolor=marker_edgecolor,
                zorder=marker_zorder,
                label='start' if ped == 0 else None)
        ax.plot(states[0, ped:ped + 1, 6], states[0, ped:ped + 1, 7],
                'x', color=marker_color,
                alpha=marker_alpha,
                markeredgewidth=2,
                zorder=marker_zorder,
                label='goal' if ped == 0 else None)


def space(ax, ped_space_potential):
    for s in ped_space_potential.space:
        s = np.asarray(s)
        ax.plot(s[:, 0], s[:, 1], '-o', color='black', markersize=2.5)


def potential_1d(V, ax1, ax2=None, **kwargs):
    parameters = list(V.parameters())
    dtype = parameters[0].dtype if parameters else torch.float32
    b = torch.linspace(0.0, 3.0, 200, dtype=dtype)

    with torch.no_grad():
        y = V.value_b(b)
        y = y - y[-1]

    ax1.set_xlabel('$b$ [m]')
    ax1.set_ylabel('$V$')
    ax1.plot(b, y, **kwargs)
    ax1.legend()

    if ax2 is not None:
        ax2.set_xlabel(r'$b$ [m]')
        ax2.set_ylabel(r'$\nabla V$')
        delta_b = b[1:] - b[:-1]
        average_b = 0.5 * (b[:-1] + b[1:])
        diff_b = y[1:] - y[:-1]
        ax2.plot(average_b, diff_b / delta_b, **kwargs)
        ax2.legend()


def potential_1d_parametric(V, ax1, ax2=None, label=None, sigma_label=None,
                            linestyle=None, **kwargs):
    potential_1d(V, ax1, ax2, linestyle=linestyle, label=label, **kwargs)
    ax1.axvline(V.sigma, linestyle='dotted', label=sigma_label, **kwargs)
    if ax2 is not None:
        ax2.axvline(V.sigma, linestyle='dotted', label=sigma_label, **kwargs)


def potential_2d(V, ax, nx=600, ny=400, **kwargs):
    # the "pedestrian of interest" is beta and the probe pedestrians are alpha

    parameters = list(V.parameters())
    dtype = parameters[0].dtype if parameters else torch.float32
    x1 = torch.linspace(-1.0, 2.0, nx, dtype=dtype)
    x2 = torch.linspace(-1.0, 1.0, ny, dtype=dtype)
    xx1, xx2 = torch.meshgrid(x1, x2)
    r_ab_probe = torch.stack((xx1, xx2), dim=-1).view(-1, 2).unsqueeze(1)

    speeds_b = torch.full((1,), 1.0, dtype=dtype)
    desired_directions_b = torch.zeros((1, 2), dtype=dtype)
    desired_directions_b[:, 0] = 1.0

    ax.axhline(0.0, ls='dotted', color='gray')
    ax.axvline(0.0, ls='dotted', color='gray')

    with torch.no_grad():
        values = V.value_r_ab(r_ab_probe, speeds_b, desired_directions_b)
        values -= torch.min(values)
        values = values.view((len(x1), len(x2)))

    ax.clabel(
        ax.contour(x1, x2, values.T,
                   levels=np.linspace(0.1, 1.5, 15), vmin=0.1, vmax=1.5, **kwargs),
        inline=1, fontsize=10)
    ax.plot([0.0], [0.0], '-', label=r'$V$', color='seagreen')  # just for legend

    ax.plot([0.0], [0.0], 'o', label='pedestrian', markersize=5.0, color='black')
    ax.arrow(0.0, 0.0, 0.4, 0.0, width=0.01, zorder=10, color='black')
    ax.set_xlabel('$x_1$ [m]')
    ax.set_ylabel('$x_2$ [m]')
    ax.set_aspect('equal')
    ax.legend()


def potential_2d_grad(V, ax, nx=600, ny=400, **kwargs):
    # the "pedestrian of interest" is beta and the probe pedestrians are alpha

    parameters = list(V.parameters())
    dtype = parameters[0].dtype if parameters else torch.float32
    x1 = torch.linspace(-1.0, 2.0, nx, dtype=dtype)
    x2 = torch.linspace(-1.0, 1.0, ny, dtype=dtype)
    xx1, xx2 = torch.meshgrid(x1, x2)
    r_ab_probe = torch.stack((xx1, xx2), dim=-1).view(-1, 2).unsqueeze(1)

    speeds_b = torch.full((1,), 1.0, dtype=dtype)
    desired_directions_b = torch.zeros((1, 2), dtype=dtype)
    desired_directions_b[:, 0] = 1.0

    ax.axhline(0.0, ls='dotted', color='gray')
    ax.axvline(0.0, ls='dotted', color='gray')

    with torch.no_grad():
        grad = V.grad_r_ab_(r_ab_probe, speeds_b, desired_directions_b)
        values = torch.linalg.norm(grad, ord=2, dim=-1)
        # values -= torch.min(values)
        values = values.view((len(x1), len(x2)))

    ax.clabel(
        ax.contour(x1, x2, values.T,
                   levels=np.linspace(0.1, 1.5, 15), vmin=0.1, vmax=1.5, **kwargs),
        inline=1, fontsize=10)
    ax.plot([0.0], [0.0], '-', label=r'$|\nabla V|$', color='seagreen')  # just for legend

    ax.plot([0.0], [0.0], 'o', label='pedestrian', markersize=5.0, color='black')
    ax.arrow(0.0, 0.0, 0.4, 0.0, width=0.01, zorder=10, color='black')
    ax.set_xlabel('$x_1$ [m]')
    ax.set_ylabel('$x_2$ [m]')
    ax.set_aspect('equal')
    ax.legend()


def experience(ax, experiences, *, reference_ped=0, **kwargs):
    ax.axhline(0.0, ls='dotted', color='gray')
    ax.axvline(0.0, ls='dotted', color='gray')

    for t0, t1 in experiences:
        for other_i in range(t0.shape[0]):
            if other_i == reference_ped:
                continue
            diff_t0 = t0[other_i] - t0[reference_ped]
            diff_t1 = t1[other_i] - t1[reference_ped]
            x = [diff_t0[0], diff_t1[0]]
            y = [diff_t0[1], diff_t1[1]]
            ax.plot(x, y, '-o', markersize=2.5, markeredgewidth=0, **kwargs)



"""Interaction potentials."""

import math
import torch

from .pedped_1d import PedPedPotential, PedPedPotentialMLP


class PedPedPotential2D(torch.nn.Module):
    """Ped-ped interaction potential based on distance b and relative angle.

    v0 is in m^2 / s^2.
    sigma is in m.
    """
    delta_t_step = PedPedPotential.delta_t_step

    b = PedPedPotential.b
    value_b = PedPedPotential.value_b
    r_ab = staticmethod(PedPedPotential.r_ab)
    norm_r_ab = staticmethod(PedPedPotential.norm_r_ab)
    grad_r_ab = PedPedPotential.grad_r_ab
    grad_r_ab_ = PedPedPotential.grad_r_ab_
    forward = PedPedPotential.forward

    def __init__(self, v0=2.1, sigma=0.3, asymmetry=0.0):
        super().__init__()

        self.v0 = v0
        self.sigma = sigma
        self.asymmetry = asymmetry
        self.register_buffer('rot90', torch.tensor([[0.0, 1.0], [-1.0, 0.0]]))

    @staticmethod
    def parallel_d(r_ab, desired_directions):
        parallel_d = torch.einsum('abj,bj->ab', r_ab, desired_directions)
        torch.diagonal(parallel_d)[:] = 0.0
        return parallel_d

    @staticmethod
    def asymmetry_factor(asymmetry, perpendicular_d):
        return 1.0 / math.log(2.0) * torch.nn.functional.softplus(asymmetry * perpendicular_d)

    def perpendicular_d(self, r_ab, desired_directions):
        desired_directions_p = torch.matmul(desired_directions, self.rot90)
        perpendicular_d = torch.einsum('abj,bj->ab', r_ab, desired_directions_p)
        torch.diagonal(perpendicular_d)[:] = 0.0
        return perpendicular_d

    def value_r_ab(self, r_ab, speeds, desired_directions):
        """Value of potential explicitely parametrized with r_ab."""
        b = self.b(r_ab, speeds, desired_directions)
        value = self.value_b(b)
        if self.asymmetry != 0.0:
            perpendicular_d = self.perpendicular_d(r_ab, desired_directions)
            value = self.asymmetry_factor(self.asymmetry, perpendicular_d) * value
        return value


class PedPedPotentialDiamond(PedPedPotential2D):
    """Ped-ped interaction potential."""
    def __init__(self, v0=2.1, sigma=0.3, *,
                 asymmetry=0.0,
                 asymmetry_offset=0.0,
                 asymmetry_angle=0.0,
                 speed_dependent=True):
        super().__init__()

        self.v0 = v0
        self.sigma = sigma
        self.asymmetry = asymmetry
        self.asymmetry_offset = asymmetry_offset
        self.speed_dependent = speed_dependent

        if asymmetry_angle:
            asymmetry_angle *= math.pi / 180.0
            self.register_buffer('asymmetry_rotation', torch.tensor([
                [math.cos(asymmetry_angle), math.sin(asymmetry_angle)],
                [-math.sin(asymmetry_angle), math.cos(asymmetry_angle)],
            ]))
        else:
            self.asymmetry_rotation = None

    def value_r_ab(self, r_ab, speeds, desired_directions):
        """Value of potential explicitely parametrized with r_ab."""
        assert speeds.requires_grad is False
        assert desired_directions.requires_grad is False

        if self.asymmetry_rotation is not None:
            desired_directions = torch.einsum(
                'ai,ij->aj', desired_directions, self.asymmetry_rotation)

        parallel_d = self.parallel_d(r_ab, desired_directions)
        perpendicular_d = self.perpendicular_d(r_ab, desired_directions)
        if self.asymmetry_offset:
            perpendicular_d += self.asymmetry_offset

        sigma_perp = self.sigma
        sigma_parallel = self.sigma
        if self.asymmetry or self.speed_dependent:
            sigma_perp = torch.full_like(perpendicular_d, self.sigma, requires_grad=False)
            sigma_parallel = torch.full_like(parallel_d, self.sigma, requires_grad=False)
            if self.asymmetry != 0.0:
                sigma_perp *= 1.0 + self.asymmetry * torch.sign(perpendicular_d)
            if self.speed_dependent:
                front = parallel_d > 0.0
                # Given speeds are for pedestrian b. Need to make it of shape ab
                # and repeat the values along the dimension for pedestrian a.
                speeds_ab = torch.repeat_interleave(
                    torch.unsqueeze(speeds, 0), sigma_parallel.shape[0], dim=0)
                sigma_parallel[front] += self.delta_t_step * speeds_ab[front]

        l1 = torch.abs(perpendicular_d) / sigma_perp + torch.abs(parallel_d) / sigma_parallel
        return self.v0 * torch.clamp_min(1.0 - 0.5 * l1, 0.0)


class PedPedPotentialMLP1p1D(PedPedPotential2D):
    """Ped-ped interaction potential."""
    def __init__(self, *, hidden_units=5):
        super().__init__()

        self.pedped_b = PedPedPotentialMLP(hidden_units=hidden_units)

        perpendicular_lin1 = torch.nn.Linear(1, hidden_units)
        perpendicular_lin2 = torch.nn.Linear(hidden_units, 1)
        self.mlp_perpendicular = torch.nn.Sequential(
            perpendicular_lin1, torch.nn.Softplus(),
            perpendicular_lin2, torch.nn.Softplus(),
        )

    def value_r_ab(self, r_ab, speeds, desired_directions):
        """Value of potential explicitely parametrized with r_ab."""
        assert speeds.requires_grad is False
        assert desired_directions.requires_grad is False
        out_b = self.pedped_b.value_r_ab(r_ab, speeds, desired_directions)

        perpendicular_d = self.perpendicular_d(r_ab, desired_directions)
        out_perpendicular = self.mlp_perpendicular(
            perpendicular_d.reshape(-1, 1)
        ).view(r_ab[:, :, 0].shape)

        out = torch.mul(out_b, out_perpendicular)
        return out


class PedPedPotentialMLP2D(PedPedPotential2D):
    """Ped-ped interaction potential."""
    def __init__(self, *,
                 hidden_units=16,
                 n_hidden_layers=1,
                 n_fourier_features=None,
                 n_spherical_features=None,
                 fourier_scale=1.0,
                 tanh_range=2.0):
        super().__init__()

        input_features = 3

        assert not (n_fourier_features and n_spherical_features)
        if n_fourier_features:
            fourier_featurizer = torch.randn((n_fourier_features // 2,))
            self.register_buffer('featurizer', fourier_featurizer)
            input_features *= n_fourier_features
        elif n_spherical_features:
            spherical_featurizer = torch.randn((input_features, n_spherical_features // 2))
            self.register_buffer('featurizer', spherical_featurizer)
            input_features = n_spherical_features
        else:
            self.featurizer = None
        self.fourier_scale = fourier_scale
        self.tanh_range = tanh_range

        lin_in = torch.nn.Linear(input_features, hidden_units)
        lin_hidden = [torch.nn.Linear(hidden_units, hidden_units)
                      for _ in range(n_hidden_layers)]
        lin_out = torch.nn.Linear(hidden_units, 1)

        # activation_function = torch.nn.Softplus
        activation_function = lambda: torch.nn.Softplus(beta=5)
        self.mlp = torch.nn.Sequential(
            lin_in, activation_function(),
            *[layer for lin in lin_hidden for layer in (lin, activation_function())],
            lin_out, torch.nn.Softplus(),
        )

    def input_features(self, r_ab, speeds, desired_directions):
        input_vector = torch.stack((
            self.b(r_ab, speeds, desired_directions),
            self.perpendicular_d(r_ab, desired_directions),
            self.parallel_d(r_ab, desired_directions),
        ), dim=-1)

        if self.tanh_range is not None:
            input_vector = self.tanh_range * torch.tanh(input_vector / self.tanh_range)
        if self.featurizer is not None:
            input_vector = self.fourier_features(input_vector)

        return input_vector

    def fourier_features(self, input_vector):
        input_vector = 2.0 * math.pi / self.fourier_scale * input_vector
        if len(self.featurizer.shape) > 1:  # spherical
            ff = torch.matmul(input_vector, self.featurizer)
        else:
            ff = torch.matmul(torch.unsqueeze(input_vector, -1),
                              torch.unsqueeze(self.featurizer, 0))
            ff = torch.reshape(ff, list(ff.shape)[:-2] + [-1])
        ff = torch.cat((torch.sin(ff), torch.cos(ff)), dim=-1)
        return ff

    def value_r_ab(self, r_ab, speeds, desired_directions):
        """Value of potential explicitely parametrized with r_ab."""
        input_vector = self.input_features(r_ab, speeds, desired_directions)
        flattened = input_vector.view(-1, input_vector.shape[-1])
        out = self.mlp(flattened).view(r_ab[:, :, 0].shape)
        out = out.clone()
        torch.diagonal(out)[:] = 0.0
        return out



# coding=utf-8

"""Synthetic pedestrian behavior according to the Social Force model.

See Helbing and Molnár 1998.
"""

import torch

from .potentials import PedPedPotential
from .field_of_view import FieldOfView
from . import stateutils


class Simulator(torch.nn.Module):
    """Simulate social force model.

    Main interface is the state. Every pedestrian is an entry in the state and
    represented by a vector (x, y, v_x, v_y, [a_x, a_y], [d_x, d_y], [tau]),
    which are the coordinates for position, velocity and destination.
    destination and tau are optional in this vector.

    ped_space is an instance of PedSpacePotential.
    ped_ped is an instance of PedPedPotential.

    delta_t in seconds.
    tau in seconds: either float or numpy array of shape[n_ped].
    field_of_view: use -1 to remove
    """
    max_speed_multiplier = 1.3

    def __init__(self, *,
                 ped_space=None, ped_ped=None, field_of_view=None,
                 delta_t=0.4, tau=0.5,
                 oversampling=10, dtype=None, integrator=None):
        super().__init__()

        self.tau = tau
        self.dtype = dtype if dtype is not None else torch.float32

        self.delta_t = delta_t / oversampling
        self.oversampling = oversampling

        if integrator is None:
            integrator = LeapfrogIntegrator(self.delta_t)
        self.integrator = integrator

        # potentials
        self.V = ped_ped or PedPedPotential()
        self.U = ped_space

        # field of view
        self.w = field_of_view or FieldOfView()

    def normalize_state(self, state):
        if not isinstance(state, torch.Tensor):
            state = torch.tensor(state, dtype=self.dtype)
        assert len(state.shape) == 2

        if state.shape[1] == 4:
            # accelerations and destinations not given
            no_destinations = torch.full((state.shape[0], 2), float('nan'),
                                         dtype=state.dtype)
            state = torch.cat((state, no_destinations), dim=-1)
        if state.shape[1] == 5:
            # accelerations and destinations not given but tau is given
            no_dest_acc = torch.full((state.shape[0], 4), float('nan'),
                                     dtype=state.dtype)
            state = torch.cat((state[:, 0:4], no_dest_acc, state[:, 4:]), dim=-1)
        if state.shape[1] in (6, 7):
            # accelerations not given
            no_accelerations = torch.zeros((state.shape[0], 2), dtype=state.dtype)
            state = torch.cat((state[:, :4], no_accelerations, state[:, 4:]), dim=-1)
        if state.shape[1] == 8:
            # tau not given
            if hasattr(self.tau, 'shape'):
                tau = self.tau
            else:
                tau = self.tau * torch.ones(state.shape[0], dtype=state.dtype)
            state = torch.cat((state, tau.unsqueeze(-1)), dim=-1)
        if state.shape[1] == 9:
            # preferred speed not given
            preferred_speeds = stateutils.speeds(state)
            state = torch.cat((state, preferred_speeds.unsqueeze(-1)), dim=-1)

        assert state.shape[1] == 10, state.shape[1]
        return state

    def f_ab(self, state):
        """Compute f_ab."""
        return -1.0 * self.V.grad_r_ab(state)

    def f_aB(self, state):
        """Compute f_aB."""
        if self.U is None:
            return None
        return -1.0 * self.U.grad_r_aB(state)

    def cap_velocity(self, state):
        """Scale down a desired velocity to its capped speed."""
        desired_velocity = state[:, 2:4]
        desired_speeds = torch.linalg.norm(desired_velocity, ord=2, dim=-1, keepdims=True)
        max_speeds = state[:, 9:10] * self.max_speed_multiplier
        factor = torch.clamp(max_speeds / desired_speeds, max=1.0)

        state = torch.clone(state)
        state[:, 2:4] = desired_velocity * factor
        return state

    def forward(self, *args):
        """Do oversampling steps."""
        state = args[0]
        state = self.normalize_state(state.clone().detach())

        for _ in range(self.oversampling):
            state = self._step(state)

        return state

    def _step(self, state):
        """Do one step in the simulation and update the state in place."""

        # accelerate to preferred velocity
        e = stateutils.desired_directions(state).detach()
        vel = state[:, 2:4].detach()
        tau = state[:, 8:9].detach()
        preferred_speeds = state[:, 9:10].detach()
        F0 = 1.0 / tau * (preferred_speeds * e - vel)

        # repulsive terms between pedestrians
        f_ab = self.f_ab(state.detach())

        # field of view modulation
        if f_ab is not None and self.w is not None and self.w != -1:
            # w = self.w(e, -f_ab).unsqueeze(-1).detach()
            r_ab = PedPedPotential.r_ab(state)
            w = self.w(e, -r_ab).unsqueeze(-1).detach()
            F_ab = w * f_ab
        else:
            F_ab = f_ab

        # repulsive terms between pedestrians and boundaries
        F_aB = self.f_aB(state.detach())

        # social force
        F = F0
        if F_ab is not None:
            F += torch.sum(F_ab, dim=1)
        if F_aB is not None:
            F += torch.sum(F_aB, dim=1)

        state = self.integrator(state, F)
        state = self.cap_velocity(state)
        return state

    def run(self, state, n_steps):
        state = self.normalize_state(state.clone().detach())
        states = [state]
        for _ in range(n_steps):
            last_state = states[-1]
            states.append(self(last_state).clone())

        return torch.stack(states)


class EulerIntegrator:
    def __init__(self, delta_t):
        self.delta_t = delta_t

    def __call__(self, state, acceleration):
        # before applying updates to state, make a copy of it
        previous_state = state
        new_state = state.clone().detach()  # gradients will be connected below

        # update state
        new_state[:, 0:2] = previous_state[:, 0:2] + self.delta_t * previous_state[:, 2:4]
        new_state[:, 2:4] = previous_state[:, 2:4] + self.delta_t * previous_state[:, 4:6]
        new_state[:, 4:6] = acceleration

        return new_state


class LeapfrogIntegrator:
    def __init__(self, delta_t):
        self.delta_t = delta_t

    def __call__(self, state, acceleration):
        # before applying updates to state, make a copy of it
        previous_state = state
        new_state = state.clone().detach()  # gradients will be connected below

        # update position
        new_state[:, 0:2] = (previous_state[:, 0:2]
                             + previous_state[:, 2:4] * self.delta_t
                             + 0.5 * acceleration * self.delta_t**2)

        # update velocity
        v = previous_state[:, 2:4] + 0.5 * (previous_state[:, 4:6] + acceleration) * self.delta_t
        new_state[:, 2:4] = v

        # update acceleration
        new_state[:, 4:6] = acceleration

        return new_state


class PeriodicBoundary:
    def __init__(self, integrator, *, x_boundary=None, y_boundary=None):
        self.integrator = integrator
        self.x_boundary = x_boundary
        self.y_boundary = y_boundary

    def __call__(self, state, acceleration):
        old_state = self.integrator(state, acceleration)

        new_state = old_state.clone()
        if self.x_boundary is not None:
            width = self.x_boundary[1] - self.x_boundary[0]
            new_state[old_state[:, 0] < self.x_boundary[0], 0] += width
            new_state[old_state[:, 0] > self.x_boundary[1], 0] -= width
        if self.y_boundary is not None:
            height = self.y_boundary[1] - self.y_boundary[0]
            new_state[old_state[:, 1] < self.y_boundary[0], 1] += height
            new_state[old_state[:, 1] > self.y_boundary[1], 1] -= height

        return new_state



"""Utility functions to process state."""

import torch


def desired_directions(state):
    """Given the current state and destination, compute desired direction."""
    destination_vectors = state[:, 6:8] - state[:, 0:2]

    # support for prediction without given destination:
    # "desired_direction" is in the direction of the current velocity
    invalid_destination = torch.isnan(destination_vectors[:, 0])
    destination_vectors[invalid_destination] = state[invalid_destination, 2:4]

    norm_factors = torch.linalg.norm(destination_vectors, ord=2, dim=-1)
    norm_factors[norm_factors == 0.0] = 1.0
    return destination_vectors / norm_factors.unsqueeze(-1)


def speeds(state):
    """Return the speeds corresponding to a given state."""
    return torch.linalg.norm(state[:, 2:4], ord=2, dim=-1)



"""PyTorch implementation of the Social Force model."""

__version__ = '0.2.0'

from .field_of_view import FieldOfView
from .trainer import Trainer
from .simulator import Simulator
from . import potentials, scenarios, show, trajnet



from collections import defaultdict, namedtuple
import itertools
import json
import random

import numpy as np


TrackRow = namedtuple('Row', ['frame', 'pedestrian', 'x', 'y', 'prediction_number', 'scene_id'])
TrackRow.__new__.__defaults__ = (None, None, None, None, None, None)
SceneRow = namedtuple('Row', ['scene', 'pedestrian', 'start', 'end', 'fps', 'tag'])
SceneRow.__new__.__defaults__ = (None, None, None, None, None, None)


class Reader:
    """Read trajnet files.

    :param scene_type: None -> numpy.array,
        'rows' -> TrackRow and SceneRow,
        'paths': grouped rows (primary pedestrian first),
        'tags': numpy.array and scene tag
    :param image_file: Associated image file of the scene
    """
    def __init__(self, input_file, scene_type=None):
        if scene_type is not None and scene_type not in {'rows', 'paths', 'tags'}:
            raise Exception('scene_type not supported')
        self.scene_type = scene_type

        self.tracks_by_frame = defaultdict(list)
        self.scenes_by_id = {}

        self.read_file(input_file)

    def read_file(self, input_file):
        with open(input_file, 'r', encoding='utf8') as f:
            for line in f:
                line = json.loads(line)

                track = line.get('track')
                if track is not None:
                    row = TrackRow(track['f'], track['p'], track['x'], track['y'],
                                   track.get('prediction_number'), track.get('scene_id'))
                    self.tracks_by_frame[row.frame].append(row)
                    continue

                scene = line.get('scene')
                if scene is not None:
                    row = SceneRow(scene['id'], scene['p'], scene['s'], scene['e'],
                                   scene.get('fps'), scene.get('tag'))
                    self.scenes_by_id[row.scene] = row

    def scenes(self, randomize=False, limit=0, ids=None, sample=None):
        scene_ids = self.scenes_by_id.keys()
        if ids is not None:
            scene_ids = ids
        if randomize:
            scene_ids = list(scene_ids)
            random.shuffle(scene_ids)
        if limit:
            scene_ids = itertools.islice(scene_ids, limit)
        if sample is not None:
            scene_ids = random.sample(scene_ids, int(len(scene_ids) * sample))
        for scene_id in scene_ids:
            yield self.scene(scene_id)

    @staticmethod
    def track_rows_to_paths(primary_pedestrian, track_rows):
        primary_path = []
        other_paths = defaultdict(list)
        for row in track_rows:
            if row.pedestrian == primary_pedestrian:
                primary_path.append(row)
                continue
            other_paths[row.pedestrian].append(row)

        return [primary_path] + list(other_paths.values())

    @staticmethod
    def paths_to_xy(paths):
        """Convert paths to numpy array with nan as blanks."""
        frames = set(r.frame for r in paths[0])
        pedestrians = set(row.pedestrian
                          for path in paths
                          for row in path if row.frame in frames)
        paths = [path for path in paths if path[0].pedestrian in pedestrians]
        frames = sorted(frames)
        pedestrians = list(pedestrians)

        frame_to_index = {frame: i for i, frame in enumerate(frames)}
        xy = np.full((len(frames), len(pedestrians), 2), np.nan)

        for ped_index, path in enumerate(paths):
            for row in path:
                if row.frame not in frame_to_index:
                    continue
                entry = xy[frame_to_index[row.frame]][ped_index]
                entry[0] = row.x
                entry[1] = row.y

        return xy

    def scene(self, scene_id):
        scene = self.scenes_by_id.get(scene_id)
        if scene is None:
            raise Exception('scene with that id not found')

        frames = range(scene.start, scene.end + 1)
        track_rows = [r
                      for frame in frames
                      for r in self.tracks_by_frame.get(frame, [])]

        # return as rows
        if self.scene_type == 'rows':
            return scene_id, scene.pedestrian, track_rows

        # return as paths
        paths = self.track_rows_to_paths(scene.pedestrian, track_rows)
        if self.scene_type == 'paths':
            return scene_id, paths

        # return with scene tag
        if self.scene_type == 'tags':
            return scene_id, scene.tag, self.paths_to_xy(paths)

        # return a numpy array
        return scene_id, self.paths_to_xy(paths)



from .pedped_1d import PedPedPotential, PedPedPotentialWall, PedPedPotentialMLP
from .pedped_2d import (
    PedPedPotential2D,
    PedPedPotentialDiamond,
    PedPedPotentialMLP1p1D,
    PedPedPotentialMLP2D,
)
from .pedspace import PedSpacePotential



import logging
import random

import torch

from .potentials import PedPedPotential

LOG = logging.getLogger(__name__)


class Trainer:
    """Trainer

    train_experience: list of state tuples (each state is a torch.Tensor)
    """

    def __init__(self, simulator, optimizer, *,
                 batch_size=1, loss=None):
        self.simulator = simulator
        self.optimizer = optimizer

        self.batch_size = batch_size
        self.loss = loss
        if loss is None:
            self.loss = torch.nn.L1Loss()

    @staticmethod
    def scenes_to_experience(scenes, radius=3.0, acc_abs=0.1):
        experience = [
            (state1, state2)
            for scene in scenes
            for state1, state2 in zip(scene[:-1], scene[1:])
        ]
        n_total = len(experience)

        def keep(state1, state2):
            valid_state1 = torch.isfinite(state1[:, 0])
            valid_state2 = torch.isfinite(state2[:, 0])

            small_distance = (
                (PedPedPotential.norm_r_ab(PedPedPotential.r_ab(state1)) < radius)
                | (PedPedPotential.norm_r_ab(PedPedPotential.r_ab(state2)) < radius)
            )
            torch.diagonal(small_distance)[:] = False
            small_distance = torch.any(small_distance, dim=-1)

            acc = (torch.abs(state1[:, 4:6]) > acc_abs) | (torch.abs(state2[:, 4:6]) > acc_abs)
            acc = torch.any(acc, dim=-1)
            # keep 10% of samples without acc:
            acc[~acc] = (torch.rand(acc[~acc].shape) < 0.1)
            acc[:] = torch.any(acc)  # symmetrize

            return valid_state1 & valid_state2 & small_distance & acc

        keep_pedestrians = [keep(state1, state2) for state1, state2 in experience]
        experience = [
            (state1[k], state2[k])
            for k, (state1, state2) in zip(keep_pedestrians, experience)
            if torch.any(k)
        ]

        LOG.info('from %d scenes, extracted %d experiences, filtered to %d',
                 len(scenes), n_total, len(experience))
        return experience

    def epoch(self, train_experience, val_experience=None):
        assert val_experience is None, 'TODO'  # TODO
        random.shuffle(train_experience)

        n_batches = 0
        epoch_loss = 0.0

        for i in range(0, len(train_experience), self.batch_size):
            data = train_experience[i:i + self.batch_size]

            loss = 0.0
            for e in data:
                Y = e[1]
                X = self.simulator(e[0])
                loss = loss + self.loss(X[:, :2], Y[:, :2])

            epoch_loss += float(loss.item())
            n_batches += 1

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

        epoch_loss /= n_batches
        return epoch_loss

    def loop(self, n_epochs, train_experience, val_experience=None, *, log_interval=1):
        for i in range(n_epochs):
            loss = self.epoch(train_experience, val_experience)
            if (i + 1) % log_interval == 0:
                print(f'epoch {i + 1}: {loss}')



from setuptools import find_packages, setup

# extract version from __init__.py
with open('socialforce/__init__.py', 'r', encoding='utf8') as f:
    VERSION_LINE = [l for l in f if l.startswith('__version__')][0]
    VERSION = VERSION_LINE.split('=')[1].strip()[1:-1]


setup(
    name='socialforce',
    version=VERSION,
    packages=find_packages(),
    license='MIT',
    description='PyTorch implementation of DeepSocialForce.',
    long_description=open('README.md', encoding='utf8').read(),
    long_description_content_type='text/markdown',
    author='Sven Kreiss',
    author_email='research@svenkreiss.com',
    url='https://github.com/svenkreiss/socialforce',

    install_requires=[
        'numpy',
        'torch',
    ],
    extras_require={
        'dev': [
            'jupyter-book',
            'matplotlib',
            'nbstripout',
            'nbval',
            'pycodestyle',
            'pylint~=2.10',
            'pytest',
        ],
        'plot': [
            'flameprof',
            'matplotlib',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ]
)
